import { BigNumber, BigNumberish, ContractReceipt, ethers, Overrides, PayableOverrides, PopulatedTransaction } from 'ethers';
import { TokenId, ChainName, ChainId, NATIVE, ParsedRelayerMessage, ParsedMessage, Context, Wormhole, RelayerAbstract } from '@wormhole-foundation/connect-sdk';
import { EvmContracts } from './contracts';
export * from './contracts';
/**
 * @category EVM
 */
export declare class EvmContext extends RelayerAbstract<ethers.ContractReceipt> {
    readonly type = Context.EVM;
    readonly contracts: EvmContracts;
    protected wormhole: Wormhole;
    constructor(wormholeInstance: Wormhole);
    getForeignAsset(tokenId: TokenId, chain: ChainName | ChainId): Promise<string | null>;
    mustGetForeignAsset(tokenId: TokenId, chain: ChainName | ChainId): Promise<string>;
    /**
     * Returns the decimals for a token
     *
     * @param tokenAddr The token address
     * @param chain The token chain name or id
     * @returns The number of token decimals
     */
    fetchTokenDecimals(tokenAddr: string, chain: ChainName | ChainId): Promise<number>;
    getNativeBalance(walletAddr: string, chain: ChainName | ChainId): Promise<BigNumber>;
    getTokenBalance(walletAddr: string, tokenId: TokenId, chain: ChainName | ChainId): Promise<BigNumber | null>;
    /**
     * Approves amount for bridge transfer. If no amount is specified, the max amount is approved
     *
     * @param chain The sending chain name or id
     * @param contractAddress The contract for which to approve (i.e. bridge or relayer)
     * @param tokenAddr The token address
     * @param amount The amount to approve
     * @param overrides Optional overrides, varies by chain
     * @throws If the token address does not exist
     */
    approve(chain: ChainName | ChainId, contractAddress: string, tokenAddr: string, amount?: BigNumberish, overrides?: PayableOverrides & {
        from?: string | Promise<string>;
    }): Promise<ethers.ContractReceipt | void>;
    /**
     * Prepare a send tx for a Token Bridge transfer (does not dispatch the transaction)
     *
     * @param token The Token Identifier (chain/address) or `'native'` if sending the native token
     * @param amount The token amount to be sent
     * @param sendingChain The source chain name or id
     * @param senderAddress The address that is dispatching the transfer
     * @param recipientChain The destination chain name or id
     * @param recipientAddress The wallet address where funds will be sent (On solana, this is the associated token account)
     * @param relayerFee A fee that would go to a relayer, if any
     * @param overrides Overrides
     * @returns The populated transaction
     */
    prepareTransfer(token: TokenId | typeof NATIVE, amount: bigint, sendingChain: ChainName | ChainId, senderAddress: string, recipientChain: ChainName | ChainId, recipientAddress: string, relayerFee?: ethers.BigNumberish, overrides?: PayableOverrides & {
        from?: string | Promise<string>;
    }): Promise<ethers.PopulatedTransaction>;
    startTransfer(token: TokenId | typeof NATIVE, amount: bigint, sendingChain: ChainName | ChainId, senderAddress: string, recipientChain: ChainName | ChainId, recipientAddress: string, relayerFee?: ethers.BigNumberish, overrides?: PayableOverrides & {
        from?: string | Promise<string>;
    }): Promise<ethers.ContractReceipt>;
    /**
     * Prepares a send tx for a Token Bridge transfer with a payload.  The payload is used to convey extra information about a transfer to be utilized in an application
     *
     * @dev This _must_ be claimed on the destination chain, see {@link Wormhole#redeem | redeem}
     *
     * @param token The Token Identifier (chain/address) or `'native'` if sending the native token
     * @param amount The token amount to be sent
     * @param sendingChain The source chain name or id
     * @param senderAddress The address that is dispatching the transfer
     * @param recipientChain The destination chain name or id
     * @param recipientAddress The wallet address where funds will be sent (On solana, this is the associated token account)
     * @param payload Arbitrary bytes that can contain any addition information about a given transfer
     * @param overrides Overrides
     * @returns The populated transaction
     */
    startTransferWithPayload(token: TokenId | typeof NATIVE, amount: bigint, sendingChain: ChainName | ChainId, senderAddress: string, recipientChain: ChainName | ChainId, recipientAddress: string, payload: Uint8Array, overrides?: PayableOverrides & {
        from?: string | Promise<string>;
    }): Promise<ethers.ContractReceipt>;
    /**
     * Prepares a send tx for a Token Bridge Relay transfer. This will automatically dispatch funds to the recipient on the destination chain.
     *
     * @param token The Token Identifier (native chain/address) or `'native'` if sending the native token
     * @param amount The token amount to be sent, as a string
     * @param toNativeToken The amount of sending token to be converted to native gas token on the destination chain
     * @param sendingChain The source chain name or id
     * @param senderAddress The address that is dispatching the transfer
     * @param recipientChain The destination chain name or id
     * @param recipientAddress The wallet address where funds will be sent (On solana, this is the associated token account)
     * @param overrides Optional overrides, varies by chain
     * @returns The populated relay transaction
     */
    prepareTransferWithRelay(token: TokenId | 'native', amount: bigint, toNativeToken: string, sendingChain: ChainName | ChainId, senderAddress: string, recipientChain: ChainName | ChainId, recipientAddress: string, overrides?: PayableOverrides & {
        from?: string | Promise<string>;
    }): Promise<ethers.PopulatedTransaction>;
    startTransferWithRelay(token: TokenId | 'native', amount: bigint, toNativeToken: string, sendingChain: ChainName | ChainId, senderAddress: string, recipientChain: ChainName | ChainId, recipientAddress: string, overrides?: PayableOverrides & {
        from?: string | Promise<string>;
    }): Promise<ethers.ContractReceipt>;
    /**
     * Prepares a redeem tx, which redeems funds for a token bridge transfer on the destination chain
     *
     * @param destChain The destination chain name or id
     * @param signedVAA The Signed VAA bytes
     * @param overrides Optional overrides, varies between chains
     * @returns The populated redeem transaction
     */
    prepareRedeem(destChain: ChainName | ChainId, signedVAA: Uint8Array, overrides?: Overrides & {
        from?: string | Promise<string>;
    }): Promise<PopulatedTransaction>;
    completeTransfer(destChain: ChainName | ChainId, signedVAA: Uint8Array, overrides?: Overrides & {
        from?: string | Promise<string>;
    }): Promise<ContractReceipt>;
    calculateMaxSwapAmount(destChain: ChainName | ChainId, tokenId: TokenId, walletAddress: string): Promise<BigNumber>;
    calculateNativeTokenAmt(destChain: ChainName | ChainId, tokenId: TokenId, amount: BigNumberish, walletAddress: string): Promise<BigNumber>;
    parseMessageFromTx(tx: string, chain: ChainName | ChainId): Promise<ParsedMessage[] | ParsedRelayerMessage[]>;
    getRelayerFee(sourceChain: ChainName | ChainId, destChain: ChainName | ChainId, tokenId: TokenId): Promise<BigNumber>;
    isTransferCompleted(destChain: ChainName | ChainId, signedVaa: string): Promise<boolean>;
    formatAddress(address: string): Uint8Array;
    parseAddress(address: ethers.utils.BytesLike): string;
    formatAssetAddress(address: string): Promise<Uint8Array>;
    parseAssetAddress(address: string): Promise<string>;
    getCurrentBlock(chain: ChainName | ChainId): Promise<number>;
}
